#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_mac.h"
#include "esp_wifi.h"
#include "esp_event.h"
#include "esp_log.h"
#include "nvs_flash.h"
#include "esp_system.h"
#include "esp_netif.h"
#include <sys/param.h>
#include "freertos/event_groups.h"

#include "lwip/err.h"
#include "lwip/sys.h"
#include "lwip/ip_addr.h"
#include "lwip/sockets.h"
#include <lwip/netdb.h>

#define EXAMPLE_ESP_WIFI_SSID       "airballtest"
#define EXAMPLE_ESP_WIFI_PASS       "airballtest"
#define EXAMPLE_MULTICAST_IPV4_ADDR "192.168.1.255"
#define UDP_PORT                    30123
#define MULTICAST_TTL               255

static const char *TAG = "wifi softAP";

static void wifi_event_handler(void* arg,
			       esp_event_base_t event_base,
			       int32_t event_id,
			       void* event_data) {
  if (event_id == WIFI_EVENT_AP_STACONNECTED) {
    wifi_event_ap_staconnected_t* event = (wifi_event_ap_staconnected_t*) event_data;
    ESP_LOGI(TAG, "station "MACSTR" join, AID=%d", MAC2STR(event->mac), event->aid);
  } else if (event_id == WIFI_EVENT_AP_STADISCONNECTED) {
    wifi_event_ap_stadisconnected_t* event = (wifi_event_ap_stadisconnected_t*) event_data;
    ESP_LOGI(TAG, "station "MACSTR" leave, AID=%d", MAC2STR(event->mac), event->aid);
  }
}

void wifi_init_softap() {
  ESP_ERROR_CHECK(esp_netif_init());
  ESP_ERROR_CHECK(esp_event_loop_create_default());
  esp_netif_t * p_netif = esp_netif_create_default_wifi_ap();

  wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
  ESP_ERROR_CHECK(esp_wifi_init(&cfg));

  ESP_ERROR_CHECK(esp_event_handler_instance_register(WIFI_EVENT,
						      ESP_EVENT_ANY_ID,
						      &wifi_event_handler,
						      NULL,
						      NULL));

  wifi_config_t wifi_config = {
    .ap = {
      .ssid = EXAMPLE_ESP_WIFI_SSID,
      .ssid_len = strlen(EXAMPLE_ESP_WIFI_SSID),
      .password = EXAMPLE_ESP_WIFI_PASS,
      .max_connection = 20,
      .authmode = WIFI_AUTH_WPA3_PSK,
      .sae_pwe_h2e = WPA3_SAE_PWE_BOTH,
      .pmf_cfg = {
	.required = true,
      },
    },
  };
  if (strlen(EXAMPLE_ESP_WIFI_PASS) == 0) {
    wifi_config.ap.authmode = WIFI_AUTH_OPEN;
  }

  ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_AP));
  ESP_ERROR_CHECK(esp_wifi_set_config(WIFI_IF_AP, &wifi_config));
  ESP_ERROR_CHECK(esp_wifi_start());
  
  esp_netif_ip_info_t ipInfo;
  IP4_ADDR(&ipInfo.ip, 192,168,1,1);
  IP4_ADDR(&ipInfo.gw, 192,168,1,1);
  IP4_ADDR(&ipInfo.netmask, 255,255,255,0);
  
  esp_netif_dhcps_stop(p_netif);
  esp_netif_set_ip_info(p_netif, &ipInfo);
  esp_netif_dhcps_start(p_netif);
}

static int create_multicast_ipv4_socket() {
  struct sockaddr_in saddr = { 0 };
  int sock = -1;
  int err = 0;

  sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if (sock < 0) {
    ESP_LOGE(TAG, "Failed to create socket. Error %d", errno);
    return -1;
  }

  // Bind the socket to any address
  saddr.sin_family = PF_INET;
  saddr.sin_port = htons(UDP_PORT);
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  err = bind(sock, (struct sockaddr *)&saddr, sizeof(struct sockaddr_in));
  if (err < 0) {
    ESP_LOGE(TAG, "Failed to bind socket. Error %d", errno);
    goto err;
  }


  // Assign multicast TTL (set separately from normal interface TTL)
  uint8_t ttl = MULTICAST_TTL;
  setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(uint8_t));
  if (err < 0) {
    ESP_LOGE(TAG, "Failed to set IP_MULTICAST_TTL. Error %d", errno);
    goto err;
  }

  // All set, socket is configured for sending and receiving
  return sock;

err:
    close(sock);
    return -1;
}

static void mcast_example_task(void *pvParameters) {
  while (1) {
    
    int sock;
    sock = create_multicast_ipv4_socket();
    
    if (sock < 0) {
      ESP_LOGE(TAG, "Failed to create IPv4 multicast socket");
      vTaskDelay(5 / portTICK_PERIOD_MS);
      continue;
    }

    int send_count = 0;

    while (true) {
      int err = 1;
  
      const char sendfmt[] = "Multicast #%d sent by ESP32\n";
      char sendbuf[48];
      char addrbuf[32] = { 0 };
      
      int len = snprintf(sendbuf, sizeof(sendbuf), sendfmt, send_count++);
      if (len > sizeof(sendbuf)) {
	ESP_LOGE(TAG, "Overflowed multicast sendfmt buffer!!");
	send_count = 0;
	err = -1;
	break;
      }
      
      struct addrinfo hints = {
	.ai_flags = AI_PASSIVE,
	.ai_socktype = SOCK_DGRAM,
      };
      
      struct addrinfo *res;
      hints.ai_family = AF_INET; // For an IPv4 socket
      err = getaddrinfo(EXAMPLE_MULTICAST_IPV4_ADDR,
			NULL,
			&hints,
			&res);
      if (err < 0) {
	ESP_LOGE(TAG, "getaddrinfo() failed for IPV4 destination address. error: %d", err);
	break;
      }
      
      if (res == 0) {
	ESP_LOGE(TAG, "getaddrinfo() did not return any addresses");
	break;
      }

      ((struct sockaddr_in *)res->ai_addr)->sin_port = htons(UDP_PORT);
      inet_ntoa_r(((struct sockaddr_in *)res->ai_addr)->sin_addr, addrbuf, sizeof(addrbuf)-1);
      ESP_LOGI(TAG, "Sending to IPV4 multicast address %s:%d...",  addrbuf, UDP_PORT);

      err = sendto(sock, sendbuf, len, 0, res->ai_addr, res->ai_addrlen);
      
      freeaddrinfo(res);
      
      if (err < 0) {
	ESP_LOGE(TAG, "IPV4 sendto failed. errno: %d", errno);
	break;
      }
      
      vTaskDelay(50 / portTICK_PERIOD_MS);
    }

    ESP_LOGE(TAG, "Shutting down socket and restarting...");
    shutdown(sock, 0);
    close(sock);
  }
}

void app_main() {
  esp_err_t ret = nvs_flash_init();
  if (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND) {
    ESP_ERROR_CHECK(nvs_flash_erase());
    ret = nvs_flash_init();
  }
  ESP_ERROR_CHECK(ret);

  ESP_LOGI(TAG, "ESP_WIFI_MODE_AP");
  wifi_init_softap();
  xTaskCreate(&mcast_example_task, "mcast_task", 4096, NULL, 5, NULL);
}

